use anyhow::{
    anyhow,
    Context,
    Result,
};
use serde::{
    Deserialize,
};
use std::collections::HashMap;

#[derive(Deserialize, Debug)]
pub struct FlakeMetadata {
    #[serde(rename = "lastModified")]
    pub last_modified: i64,
    pub locked: FlakeLockedInfo,
    pub locks: FlakeLocks,
    pub original: FlakeSource,
    pub path: String,
    pub resolved: FlakeSource,
    pub revision: String,
}

#[derive(Deserialize, Debug, Clone)]
pub struct FlakeSource {
    pub owner: Option<String>,
    pub repo: Option<String>,
    pub r#ref: Option<String>,
    pub rev: Option<String>,
    pub r#type: String,
    pub url: Option<String>,
}

pub fn assemble_flake_git_uri(flake_source: &FlakeSource) -> Result<String> {
    let mut out = String::new();
    out.push_str("git+");
    out.push_str(&flake_source.url.clone()
        .context("Flake git uri does not contain an url")?);

    let mut query: Vec<String> = Vec::new();

    if let Some(r#ref) = &flake_source.r#ref {
        query.push(format!("ref={}", r#ref));
    }

    if let Some(rev) = &flake_source.rev {
        query.push(format!("rev={}", rev));
    }

    if query.len() > 0 {
        out.push_str("?");
        out.push_str(&query.join("&"));
    }

    Ok(out)
}

pub fn assemble_flake_github_uri(flake_source: &FlakeSource) -> Result<String> {
    let mut out = String::new();
    out.push_str("github:");
    out.push_str(&flake_source.owner.clone()
        .context("Flake github uri does not contain an owner")?);
    out.push_str("/");
    out.push_str(&flake_source.repo.clone()
        .context("Flake github uri does not contain an repo")?);

    if let Some(rev) = &flake_source.rev {
        out.push_str("/");
        out.push_str(&rev);
    }

    Ok(out)
}

pub fn assemble_flake_tarball_uri(flake_source: &FlakeSource) -> Result<String> {
    let mut out = String::new();
    out.push_str(&flake_source.url.clone()
        .context("Flake tarball uri does not contain an url")?);

    if let Some(rev) = &flake_source.rev {
        out.push_str("?rev=");
        out.push_str(rev);
    }

    Ok(out)
}

pub trait FlakeUri {
    fn flake_uri(&self) -> Result<String>;
}

impl FlakeUri for FlakeSource {
    fn flake_uri(&self) -> Result<String> {
        match self.r#type.as_str() {
            "git" => assemble_flake_git_uri(self),
            "github" => assemble_flake_github_uri(self),
            "tarball" => assemble_flake_tarball_uri(self),
            other => Err(anyhow!("Unsupported flake uri type {}", other)),
        }
    }
}

#[derive(Deserialize, Debug, Clone)]
pub struct FlakeLockedInfo {
    #[serde(rename = "lastModified")]
    pub last_modified: i64,
    #[serde(rename = "narHash")]
    pub nar_hash: String,
    pub owner: Option<String>,
    pub repo: Option<String>,
    pub r#ref: Option<String>,
    pub rev: Option<String>,
    pub r#type: String,
    pub url: Option<String>,
}

impl FlakeUri for FlakeLockedInfo {
    fn flake_uri(&self) -> Result<String> {
        Into::<FlakeSource>::into(self).flake_uri()
    }
}

impl Into<FlakeSource> for &FlakeLockedInfo {
    fn into(self) -> FlakeSource {
        FlakeSource {
            owner: self.owner.clone(),
            repo: self.repo.clone(),
            r#ref: self.r#ref.clone(),
            rev: self.rev.clone(),
            r#type: self.r#type.clone(),
            url: self.url.clone(),
        }
    }
}

#[derive(Deserialize, Debug)]
pub struct FlakeLocks {
    pub nodes: HashMap<String, FlakeLocksNode>,
    pub root: String,
    pub version: u64,
}

#[derive(Deserialize, Debug, Clone)]
#[serde(untagged)]
pub enum FlakeLocksNodeInputs {
    String(String),
    Array(Vec<String>),
}

#[derive(Deserialize, Debug)]
pub struct FlakeLocksNode {
    pub inputs: Option<HashMap<String, FlakeLocksNodeInputs>>,
    pub locked: Option<FlakeLockedInfo>,
    pub original: Option<FlakeSource>,
}

