use anyhow::{
    Context,
    Result,
};
use chrono::{
    Utc,
};
use crate::{
    flake::{
        FlakeLocksNodeInputs,
        FlakeMetadata,
        FlakeUri,
    },
    storage::{
        InputRow,
        FlakeRow,
        RevisionRow,
        Storage,
    },
};
use std::process::Command;

pub async fn scan_flake(storage: Storage, flake_uri: &str) -> Result<()> {
    let scan_time = Utc::now().timestamp();

    let flake_metadata = fetch_metadata(flake_uri)
        .await?;

    let revision_row = get_revision_from_metadata(&flake_metadata)?;

    storage.set_revision(revision_row)
        .await?;

    let mut flake_row = get_flake_from_metadata(&flake_metadata)?;
    flake_row.tracker_last_scanned = Some(scan_time.clone());

    storage.set_flake(flake_row)
        .await?;

    let locks_root_name = &flake_metadata.locks.root;
    let locks_root_node = flake_metadata.locks.nodes.get(locks_root_name)
        .context("Failed to get locks root node")?;

    for (input_name, locks_input_name) in locks_root_node.inputs.clone().context("No inputs found for flake")? {

        if let FlakeLocksNodeInputs::String(locks_input_name) = locks_input_name {
            let locks_input_node = flake_metadata.locks.nodes.get(&locks_input_name)
                .context("Failed to find lock of input")?;

            let input_row = InputRow {
                revision_uri: flake_metadata.locked.flake_uri()?.clone(),
                input_name: input_name.clone(),
                locked_revision_uri: Some(locks_input_node.locked.clone().context("Unexpected missing lock")?.flake_uri()?),
                locked_flake_uri: Some(locks_input_node.original.clone().context("Unexpected missing lock")?.flake_uri()?),
                locked_nar_hash: Some(locks_input_node.locked.clone().context("Unexpected missing lock")?.nar_hash),
                last_modified: Some(locks_input_node.locked.clone().context("Unexpected missing lock")?.last_modified),
            };
            storage.set_input(input_row)
                .await?;

            let revision_row = RevisionRow {
                revision_uri: locks_input_node.locked.clone().context("Unexpected missing lock")?.flake_uri()?.clone(),
                flake_uri: Some(locks_input_node.original.clone().context("Unexpected missing lock")?.flake_uri()?.clone()),
                nix_store_path: None,
                nar_hash: None,
                last_modified: None,
            };
            storage.set_revision_exist(revision_row)
                .await?;

            let flake_row = FlakeRow {
                flake_uri: locks_input_node.original.clone().context("Unexpected missing lock")?.flake_uri()?.clone(),
                tracker_track: false,
                tracker_last_scanned: None,
            };

            storage.set_flake_exist(flake_row)
                .await?;

        }
    }

    Ok(())
}

pub async fn fetch_metadata(flake_uri: &str) -> Result<FlakeMetadata> {
    let flake_metadata_raw = Command::new("nix")
        .arg("flake")
        .arg("metadata")
        .arg("--json")
        .arg(flake_uri)
        .output()
        .context("Failed to fetch flake metadata")?;

    let flake_metadata: FlakeMetadata = serde_json::from_slice(&flake_metadata_raw.stdout)
        .context("Failed to parse flake metadata")?;

    Ok(flake_metadata)
}

pub fn get_flake_from_metadata(flake_metadata: &FlakeMetadata) -> Result<FlakeRow> {
    let flake_row = FlakeRow {
        flake_uri: flake_metadata.resolved.flake_uri()?.clone(),
        tracker_track: false,
        tracker_last_scanned: None,
    };

    Ok(flake_row)
}

pub fn get_revision_from_metadata(flake_metadata: &FlakeMetadata) -> Result<RevisionRow> {
    let revision_row = RevisionRow {
        revision_uri: flake_metadata.locked.flake_uri()?.clone(),
        flake_uri: Some(flake_metadata.resolved.flake_uri()?.clone()),
        nix_store_path: Some(flake_metadata.path.clone()),
        nar_hash: Some(flake_metadata.locked.nar_hash.clone()),
        last_modified: Some(flake_metadata.locked.last_modified.clone()),
    };

    Ok(revision_row)
}
